{ MD5 checksum calculation for Delphi (all Versions including 1.0)
  Copyright 2007 Holger Guenther (www.hguent.de)
  This unit is based on the MD5 C code by Markus Hahn
  http://come.to/hahn , markus_hahn@gmx.net }

unit MD5;

interface

const
  MD5Size = 16;                  { size of an MD5 check sum in bytes }

type
  TMD5       = packed array[0..MD5Size-1] of byte;
  TMD5State  = packed array[0..3] of longint;
  TMD5Buffer = packed array[0..63] of byte;

  { Simulation of QuadWord for compilers not providing an int64 type }
  TMD5Count  = packed record case boolean of
    false: (l,h: longint);
    true : (q: comp);
  end;

  { MD5 context }
  PMD5Handle = ^TMD5Handle;
  TMD5Handle = record
    State: TMD5State;              { state (ABCD) }
    Count: TMD5Count;              { number of bits, modulo 2^64 (lsb first) }
    Buf  : TMD5Buffer;             { input buffer }
  end;

{ *** Low-Level API routines *** }

{ Initializes a new MD5 calculation }
function  MD5Initialize: TMD5Handle;
{ Continues the MD5 calculation with the given data block }
procedure MD5Update (var HMD5: TMD5Handle; const Data; Size: longint);
{ Finalizes the MD5 calculation and returns the result }
function  MD5Finalize (var HMD5: TMD5Handle): TMD5;

{ *** High-Level API routines *** }

{ Calculates the MD5 checksum for a memory data block }
function  MD5CalcMem (const Data; Size: longint): TMD5;

{ Calculates the MD5 checksum for a file
  Attention: Due to the use of standard pascal file routines,
  this will ONLY work for files smaller than 2 GB!
  For larger files, write your own routine using the OS file routines. }
function  MD5CalcFile (const Filename: string; var CheckSum: TMD5): boolean;

{ *** Service routines *** }

{ Returns TRUE if the two MD5 check sums are equal }
function  MD5Equal (const da,db: TMD5): boolean;

{ Converts an MD5 check sum to a hexadecimal string
 NumSpaces: Count of space chars between then hex bytes }
function  MD5AsString (const AnMD5: TMD5;
  UpperCaseChars: boolean; NumSpaces: byte): string;

{ Self-Test; returns TRUE if successful }
function  MD5SelfTest: boolean;

implementation

{ constants for MD5Transform routine }
const
  S11 =  7;         S12 = 12;         S13 = 17;         S14 = 22;
  S21 =  5;         S22 =  9;         S23 = 14;         S24 = 20;
  S31 =  4;         S32 = 11;         S33 = 16;         S34 = 23;
  S41 =  6;         S42 = 10;         S43 = 15;         S44 = 21;

type
  TByteArray = packed array[0..0] of byte;

{ F, G, H and I are basic MD5 functions }
function F (x,y,z: longint): longint;
begin
  F:=(x and y) or (z and not x);
end;

function G (x,y,z: longint): longint;
begin
  G:=(x and z) or (y and not z);
end;

function H (x,y,z: longint): longint;
begin
  H:=x xor y xor z;
end;

function I (x,y,z: longint): longint;
begin
  I:=y xor (x or not z);
end;

{ RotL rotates x left n bits }
function RotL (x,n: longint): longint;
begin
  RotL:=(x shl n) or (x shr (32-n));
end;

{ FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
  Rotation is separate from addition to prevent recomputation. }
procedure FF (var a,b,c,d: longint; x,s,ac: longint);
begin
  Inc (a,F (b,c,d)+x+ac);
  a:=RotL (a,s);
  Inc (a,b);
end;

procedure GG (var a,b,c,d: longint; x,s,ac: longint);
begin
  Inc (a,G (b,c,d)+x+ac);
  a:=RotL (a,s);
  Inc (a,b);
end;

procedure HH (var a,b,c,d: longint; x,s,ac: longint);
begin
  Inc (a,H (b,c,d)+x+ac);
  a:=RotL (a,s);
  Inc (a,b);
end;

procedure II (var a,b,c,d: longint; x,s,ac: longint);
begin
  Inc (a,I (b,c,d)+x+ac);
  a:=RotL (a,s);
  Inc (a,b);
end;

{ basic transformation, transforms state based on block }
procedure MD5Transform (var State: TMD5State; const Block);
var a,b,c,d: longint;
    x      : packed array[0..15] of longint absolute Block;
begin
  a:=State[0];
  b:=State[1];
  c:=State[2];
  d:=State[3];
  { round 1 }
  FF ( a, b, c, d, x[ 0], S11, $d76aa478);   {  1 }
  FF ( d, a, b, c, x[ 1], S12, $e8c7b756);   {  2 }
  FF ( c, d, a, b, x[ 2], S13, $242070db);   {  3 }
  FF ( b, c, d, a, x[ 3], S14, $c1bdceee);   {  4 }
  FF ( a, b, c, d, x[ 4], S11, $f57c0faf);   {  5 }
  FF ( d, a, b, c, x[ 5], S12, $4787c62a);   {  6 }
  FF ( c, d, a, b, x[ 6], S13, $a8304613);   {  7 }
  FF ( b, c, d, a, x[ 7], S14, $fd469501);   {  8 }
  FF ( a, b, c, d, x[ 8], S11, $698098d8);   {  9 }
  FF ( d, a, b, c, x[ 9], S12, $8b44f7af);   { 10 }
  FF ( c, d, a, b, x[10], S13, $ffff5bb1);   { 11 }
  FF ( b, c, d, a, x[11], S14, $895cd7be);   { 12 }
  FF ( a, b, c, d, x[12], S11, $6b901122);   { 13 }
  FF ( d, a, b, c, x[13], S12, $fd987193);   { 14 }
  FF ( c, d, a, b, x[14], S13, $a679438e);   { 15 }
  FF ( b, c, d, a, x[15], S14, $49b40821);   { 16 }
  { round 2 }
  GG ( a, b, c, d, x[ 1], S21, $f61e2562);   { 17 }
  GG ( d, a, b, c, x[ 6], S22, $c040b340);   { 18 }
  GG ( c, d, a, b, x[11], S23, $265e5a51);   { 19 }
  GG ( b, c, d, a, x[ 0], S24, $e9b6c7aa);   { 20 }
  GG ( a, b, c, d, x[ 5], S21, $d62f105d);   { 21 }
  GG ( d, a, b, c, x[10], S22,  $2441453);   { 22 }
  GG ( c, d, a, b, x[15], S23, $d8a1e681);   { 23 }
  GG ( b, c, d, a, x[ 4], S24, $e7d3fbc8);   { 24 }
  GG ( a, b, c, d, x[ 9], S21, $21e1cde6);   { 25 }
  GG ( d, a, b, c, x[14], S22, $c33707d6);   { 26 }
  GG ( c, d, a, b, x[ 3], S23, $f4d50d87);   { 27 }
  GG ( b, c, d, a, x[ 8], S24, $455a14ed);   { 28 }
  GG ( a, b, c, d, x[13], S21, $a9e3e905);   { 29 }
  GG ( d, a, b, c, x[ 2], S22, $fcefa3f8);   { 30 }
  GG ( c, d, a, b, x[ 7], S23, $676f02d9);   { 31 }
  GG ( b, c, d, a, x[12], S24, $8d2a4c8a);   { 32 }
  { round 3 }
  HH ( a, b, c, d, x[ 5], S31, $fffa3942);   { 33 }
  HH ( d, a, b, c, x[ 8], S32, $8771f681);   { 34 }
  HH ( c, d, a, b, x[11], S33, $6d9d6122);   { 35 }
  HH ( b, c, d, a, x[14], S34, $fde5380c);   { 36 }
  HH ( a, b, c, d, x[ 1], S31, $a4beea44);   { 37 }
  HH ( d, a, b, c, x[ 4], S32, $4bdecfa9);   { 38 }
  HH ( c, d, a, b, x[ 7], S33, $f6bb4b60);   { 39 }
  HH ( b, c, d, a, x[10], S34, $bebfbc70);   { 40 }
  HH ( a, b, c, d, x[13], S31, $289b7ec6);   { 41 }
  HH ( d, a, b, c, x[ 0], S32, $eaa127fa);   { 42 }
  HH ( c, d, a, b, x[ 3], S33, $d4ef3085);   { 43 }
  HH ( b, c, d, a, x[ 6], S34,  $4881d05);   { 44 }
  HH ( a, b, c, d, x[ 9], S31, $d9d4d039);   { 45 }
  HH ( d, a, b, c, x[12], S32, $e6db99e5);   { 46 }
  HH ( c, d, a, b, x[15], S33, $1fa27cf8);   { 47 }
  HH ( b, c, d, a, x[ 2], S34, $c4ac5665);   { 48 }
  { round 4 }
  II ( a, b, c, d, x[ 0], S41, $f4292244);   { 49 }
  II ( d, a, b, c, x[ 7], S42, $432aff97);   { 50 }
  II ( c, d, a, b, x[14], S43, $ab9423a7);   { 51 }
  II ( b, c, d, a, x[ 5], S44, $fc93a039);   { 52 }
  II ( a, b, c, d, x[12], S41, $655b59c3);   { 53 }
  II ( d, a, b, c, x[ 3], S42, $8f0ccc92);   { 54 }
  II ( c, d, a, b, x[10], S43, $ffeff47d);   { 55 }
  II ( b, c, d, a, x[ 1], S44, $85845dd1);   { 56 }
  II ( a, b, c, d, x[ 8], S41, $6fa87e4f);   { 57 }
  II ( d, a, b, c, x[15], S42, $fe2ce6e0);   { 58 }
  II ( c, d, a, b, x[ 6], S43, $a3014314);   { 59 }
  II ( b, c, d, a, x[13], S44, $4e0811a1);   { 60 }
  II ( a, b, c, d, x[ 4], S41, $f7537e82);   { 61 }
  II ( d, a, b, c, x[11], S42, $bd3af235);   { 62 }
  II ( c, d, a, b, x[ 2], S43, $2ad7d2bb);   { 63 }
  II ( b, c, d, a, x[ 9], S44, $eb86d391);   { 64 }
  Inc (State[0],a);
  Inc (State[1],b);
  Inc (State[2],c);
  Inc (State[3],d);
end;

{ *** API *** }

{ Initializes a new MD5 calculation }
function MD5Initialize: TMD5Handle;
var Handle: TMD5Handle;
begin
  FillChar (Handle,SizeOf (Handle),0);
  { load magic init. constants }
  Handle.State[0]:=$67452301;
  Handle.State[1]:=$efcdab89;
  Handle.State[2]:=$98badcfe;
  Handle.State[3]:=$10325476;
  MD5Initialize:=Handle;
end;

{ Continues the MD5 calculation with the given data block }
procedure MD5Update (var HMD5: TMD5Handle; const Data; Size: longint);
var Inp             : TByteArray absolute Data;
    p,Index,PartLen : longint;
begin
  with HMD5 do begin
    { compute number of bytes mod 64 }
    Index:=(Count.l shr 3) and $3F;
    { update number of bits }
    Count.q:=Count.q+Size shl 3;
    PartLen:=64-Index;
    { transform as many times as possible }
    if Size>=PartLen then begin
      Move (Inp[0],Buf[Index],PartLen);
      MD5Transform (State,Buf);
      p:=PartLen;
      while p+63<Size do begin
        MD5Transform (State,Inp[p]);
        Inc (p,64);
      end;
      Index:=0;
      end
    else p:=0;
    { buffer remaining input }
    if p<Size then Move (Inp[p],Buf[Index],Size-p);
  end;
end;

{ Finalizes the MD5 calculation and returns the result }
function MD5Finalize (var HMD5: TMD5Handle): TMD5;
var Padding: TMD5Buffer;
    Bits   : packed array[0..7] of byte;
    Index  : longint;
    PadLen : longint;
    Res    : TMD5;
begin
  FillChar (Padding,SizeOf (Padding),0);
  Padding[0]:=$80;
  { save number of bits }
  Move (HMD5.Count,Bits,SizeOf (Bits));
  { pad out to 56 mod 64 }
  Index:=(HMD5.Count.l shr 3) and $3F;
  if Index<56 then PadLen:=56-Index else PadLen:=120-Index;
  MD5Update (HMD5,Padding,PadLen);
  { append length (before padding) }
  MD5Update (HMD5,Bits,8);
  { store state to result }
  Move (HMD5.State,Res,SizeOf (Res));
  { clear context }
  FillChar (HMD5,SizeOf (HMD5),0);
  MD5Finalize:=Res;
end;

{ Calculates the MD5 checksum for a memory data block }
function MD5CalcMem (const Data; Size: longint): TMD5;
var Handle: TMD5Handle;
begin
  Handle:=MD5Initialize;
  MD5Update (Handle,Data,Size);
  MD5CalcMem:=MD5Finalize (Handle);
end;

{ Calculates the MD5 checksum for a file
  Attention: Due to the use of standard pascal file routines,
  this will ONLY work for files smaller than 2 GB!
  For larger files, write your own routine using the OS file routines. }
function MD5CalcFile (const Filename: string; var CheckSum: TMD5): boolean;
const BUFSIZE = {$IFDEF VER80} 4 {$ELSE} 64 {$ENDIF} *1024;
var f           : file;
    OldFileMode : byte;
    FSize       : longint;
    BytesToRead : integer;
    ReallyRead  : integer;
    Buffer      : packed array[0..BUFSIZE-1] of byte;
    Handle      : TMD5Handle;
    ok          : boolean;
begin
  Handle:=MD5Initialize;
  OldFileMode:=FileMode;
  FileMode:=0;
  AssignFile (f,Filename);
  {$IFDEF VER80} ok:=IOResult=0 {$ELSE} IOResult {$ENDIF} ;
  Reset (f,1);
  ok:=IOResult=0;
  if ok then begin
    FSize:=FileSize (f);
    while FSize>0 do begin
      if FSize>BUFSIZE then BytesToRead:=BUFSIZE else BytesToRead:=FSize;
      BlockRead (f,Buffer,BytesToRead,ReallyRead);
      ok:=BytesToRead=ReallyRead;
      if not ok then break;
      MD5Update (Handle,Buffer,ReallyRead);
      Dec (FSize,ReallyRead);
    end;
  end;
  CheckSum:=MD5Finalize (Handle);
  CloseFile (f);
  FileMode:=OldFileMode;
  MD5CalcFile:=ok;
end;

{ Returns TRUE if the two MD5 check sums are equal }
function MD5Equal (const da,db: TMD5): boolean;
var i: byte;
begin
  MD5Equal:=false;
  for i:=0 to MD5Size-1 do if da[i]<>db[i] then exit;
  MD5Equal:=true;
end;

{ Converts an MD5 check sum to a hexadecimal string
  NumSpaces: Count of space chars between then hex bytes }
function MD5AsString (const AnMD5: TMD5;
  UpperCaseChars: boolean; NumSpaces: byte): string;
var i,j  : byte;
    s,Hex: string;
begin
  if UpperCaseChars then Hex:='0123456789ABCDEF' else Hex:='0123456789abcdef';
  s:='';
  for i:=0 to MD5Size-1 do begin
    if i>0 then if NumSpaces>0 then for j:=0 to NumSpaces-1 do s:=s+' ';
    s:=s+Hex[((AnMD5[i] shr 4) and $0F)+1]+Hex[(AnMD5[i] and $0F)+1];
  end;
  MD5AsString:=s;
end;

{ the selftest strings and check sum values }
const
  SelfTestCount = 3;
  SelfTestSource: array[0..SelfTestCount-1] of string = (
    'hellooo nurse!',
    'whoa nelly, says Sherman, the Shark',
    '01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
  );
  SelfTestCheckSum: array[0..SelfTestCount-1] of TMD5 = (
    ( $13, $12, $04, $84, $56, $62, $22, $df,
      $cf, $ac, $4f, $d4, $1e, $78, $18, $90 ),
    ( $1f, $6a, $8e, $65, $ba, $18, $b7, $74,
      $c0, $c1, $ed, $cb, $e6, $29, $b4, $95 ),
    ( $9c, $dc, $a7, $fa, $46, $c7, $c7, $1a,
      $5c, $fe, $d3, $bc, $99, $9e, $2f, $91 )
  );

{ Self-Test; returns TRUE if successful }
function MD5SelfTest: boolean;
var i        : longint;
    CheckSum : TMD5;
begin
  MD5SelfTest:=false;
  for i:=0 to SelfTestCount-1 do begin
    CheckSum:=MD5CalcMem (SelfTestSource[i][1],Length (SelfTestSource[i]));
    if not MD5Equal (CheckSum,SelfTestCheckSum[i]) then exit;
  end;
  MD5SelfTest:=true;
end;

end.

